"""
Import of  default state and extras
"""

__all__ = []

# Redis edition extra
try:
    import redis.asyncio as redis
    from .redis import RedisAntenna

    __all__ += "RedisAntenna"
except ImportError as error:  # pragma: no cover
    pass



from . import antenna_implementations as antennas
from .core.interfaces import Antenna as IAntenna
from .domain.request import Request
from .domain.response import Response
from .server import XRaptor
from .broadcaster import Broadcast

__all__ = ["XRaptor", "antennas", "Request", "Response", "IAntenna", "Broadcast"]



import json
from dataclasses import dataclass
from uuid import uuid4

from xraptor.domain.methods import MethodType


@dataclass(slots=True, frozen=True)
class Request:
    request_id: str
    payload: str
    header: dict
    route: str
    method: MethodType

    def __post_init__(self):
        assert isinstance(self.request_id, str), "request_id is not of type {}".format(
            str
        )
        assert isinstance(self.payload, str), "header is not of type {}".format(str)
        assert isinstance(self.header, dict), "payload is not of type {}".format(dict)
        assert isinstance(self.route, str), "payload is not of type {}".format(str)
        assert isinstance(self.method, MethodType), "payload is not of type {}".format(
            MethodType
        )

    @classmethod
    def from_message(cls, message: str | bytes):
        """
        cast string message to a valid Request object instance
        :param message: json like string
        :return: Request instance
        """

        if isinstance(message, bytes):
            message: str = message.decode()

        message_data = json.loads(message)

        payload = message_data["payload"]

        if isinstance(payload, dict):
            payload = json.dumps(payload)

        return cls(
            request_id=message_data.get("request_id", str(uuid4())),
            payload=payload,
            header=message_data["header"],
            route=message_data["route"],
            method=MethodType[message_data["method"]],
        )



import asyncio
from typing import Self, Callable, Type, Awaitable

import witch_doctor
from websockets import serve

from xraptor import antenna_implementations as antennas
from xraptor.core.interfaces import Antenna
from xraptor.domain.methods import MethodType
from xraptor.domain.response import Response
from xraptor.domain.route import Route
from xraptor.handler import Handler


class XRaptor:
    _routes: list[Route] = []
    _map: dict = {}
    _antenna_cls: Type[object] = None

    def __init__(self, ip: str, port: int):
        self._ip = ip
        self._port = port
        self._server = None

    @classmethod
    def set_antenna(cls, antenna: Type[Antenna]):
        """
        set new antenna implementation
        :param antenna: class that implements all Antenna methods
        :return:
        """
        cls._antenna_cls = antenna

    @classmethod
    def _load_oic(cls):
        """
        load oic container with the registered antenna implementation
        :return:
        """
        if cls._antenna_cls is None:
            cls._antenna_cls = antennas.RedisAntenna

        assert issubclass(
            cls._antenna_cls, Antenna
        ), "antenna is not subtype of {}".format(Antenna)

        container = witch_doctor.WitchDoctor.container()
        container(
            Antenna,
            cls._antenna_cls,
            witch_doctor.InjectionType.FACTORY,
        )
        witch_doctor.WitchDoctor.load_container()

    def load_routes(self) -> Self:
        """
        load all registered routes on server
        :return:
        """
        [self._map.update(r.get_match_map()) for r in self._routes]
        self._load_oic()
        return self

    async def serve(self):
        """
        start serve
        :return:
        """
        async with serve(Handler.watch, self._ip, self._port) as server:
            self._server = server
            while True:
                await asyncio.sleep(10)

    @classmethod
    def register(cls, name: str) -> Route:
        """
        register a route by name and return a Route instance that allow you to register as one of possible route types
        :param name: route name
        :return:
        """
        _route = Route(name)
        cls._routes.append(_route)
        return _route

    @classmethod
    def route_matcher(
        cls, method: MethodType, name: str
    ) -> Callable[..., Awaitable[Response | None]] | None:
        """
        will return the registered async callback for the giving method and route name if registered
        :param method: on of the allowed MethodType
        :param name: route name
        :return:
        """
        key = f"{name}:{method.value}"
        return cls._map.get(key)



import asyncio
import logging
from typing import Callable, Awaitable

import websockets
from websockets.frames import CloseCode

import xraptor
from xraptor.connection import Connection
from xraptor.domain.methods import MethodType
from xraptor.domain.request import Request
from xraptor.domain.response import Response


class Handler:

    @staticmethod
    async def watch(websocket: websockets.WebSocketServerProtocol):
        connection = Connection.from_ws(websocket)
        close_code: CloseCode = CloseCode.NORMAL_CLOSURE
        try:
            async for message in connection.ws:
                await Handler._handle_request(message, connection)
        except websockets.exceptions.ConnectionClosed as e:
            logging.error(e)
            close_code = CloseCode.GOING_AWAY
        except websockets.exceptions.InvalidHandshake as e:
            logging.error(e)
            close_code = CloseCode.TLS_HANDSHAKE
        except websockets.exceptions.WebSocketException as e:
            logging.error(e)
            close_code = CloseCode.PROTOCOL_ERROR
        except Exception as e:
            logging.error(e)
            close_code = CloseCode.ABNORMAL_CLOSURE
        finally:
            await connection.close(close_code=close_code)
            del connection

    @staticmethod
    async def _handle_request(message: str | bytes, connection: Connection):
        try:
            request = Request.from_message(message)
        except AssertionError:
            return

        try:
            result = None
            if fn := xraptor.XRaptor.route_matcher(request.method, request.route):
                if (
                    request.method == MethodType.GET
                    or request.method == MethodType.POST
                ):
                    result = await fn(request)
                if request.method == MethodType.SUB:
                    result = await Handler._subscribe(request, connection, fn)
                if request.method == MethodType.UNSUB:
                    result = await fn(request)
                    connection.unregister_response_receiver(request)

                if result is not None:
                    await connection.ws.send(result.json())
                return
            await connection.ws.send(
                Response.create(
                    request_id=request.request_id,
                    header={},
                    payload='{"message": "Not registered"}',
                ).json()
            )
        except Exception as e:
            logging.error(e)
            _response = Response.create(
                request_id=request.request_id, header={}, payload='{"message": "fail"}'
            )
            await connection.ws.send(_response.json())

    @staticmethod
    async def _subscribe(
        request: Request, connection: Connection, fn: Callable
    ) -> Awaitable[Response | None]:
        try:
            connection.register_response_receiver(request)
            result = await fn(request)
            return result
        except Exception as e:
            logging.error(e)
            connection.unregister_response_receiver(request)
